from dataclasses import dataclass
from typing import Any, Callable
from .twitch import Send

# Data Classes
@dataclass
class CommandInfo:
    command: str
    args: list[str]
    user: str
    
    def __iter__(self):
        yield self.command
        yield self.args
        yield self.user

@dataclass
class CommandData:
    irc: Any
    channel: str
    command_info: CommandInfo
    
    def __iter__(self):
        yield self.irc
        yield self.channel
        yield self.command_info

# Main function
def cmd_handler(
        data: CommandData,
        COMMANDS: dict[str, Callable[[list[str], str], Any]],
        handle: str
    ) -> bool:
    """
    Handles every user message sent.
    
    Args:
        data: (irc, channel, command_info) — where command_info is (command, args, user).
        commands: Mapping of command names to callables.
    """
    
    irc = data.irc
    channel = data.channel
    info = data.command_info

    command = info.command
    args = info.args
    user = info.user

    if not command.startswith(handle):
        return False

    cmd = command[len(handle):].lower()
    
    if cmd in COMMANDS:
        try:
            result, shutdown = COMMANDS[cmd](args, user)
            if result is not None:
                Send(
                    irc=irc,
                    CHANNEL=channel,
                    MESSAGE=result
                )
                return shutdown
        except Exception as e:
            print(f"Error executing command '{cmd}': {e}")
    
    return False


def Send(irc, CHANNEL: str, MESSAGE: str):
    """
    Sends a message to the specified IRC Channel.
    
    Args:
        irc (socket): The active IRC Socket connection
        CHANNEL (str): The Twitch channel to send the message to (Must start with '#')
        MESSAGE (str): The message content
    """
    if not CHANNEL.startswith("#"):
        CHANNEL = f"#{CHANNEL}"
    
    if hasattr(irc, "send_raw"):
        irc.send_raw(f"PRIVMSG {CHANNEL} :{MESSAGE}")
    else:
        irc.send(f"PRIVMSG {CHANNEL} :{MESSAGE}\r\n".encode("utf-8"))

def GetUsername(raw: str) -> str:
    """
    Extracts the sender's username from the raw IRC message.
    
    Args:
        raw (str): Raw IRC PRIVMSG string
    """
    return raw.split("!", 1)[0][1:]


import os
import requests

from dotenv import load_dotenv

class Auth:
    def __init__(self, ENV_PATH = None):
        if ENV_PATH is None:
            raise ValueError("ENV_PATH is required but was not provided.")
        
        load_dotenv(ENV_PATH)
        self.env_path = ENV_PATH
        
        self.CLIENT_ID = os.getenv('CLIENT_ID')
        self.CLIENT_SECRET = os.getenv('CLIENT_SECRET')
        self.REFRESH_TOKEN = os.getenv('TWITCH_REFRESH_TOKEN')
        self.OAuth_TOKEN = None
        
        self.refresh_access_token()
    
    def refresh_access_token(self):
        payload = {
            'client_id': self.CLIENT_ID,
            'client_secret': self.CLIENT_SECRET,
            'refresh_token': self.REFRESH_TOKEN,
            'grant_type': 'refresh_token'
        }
        
        response = requests.post('https://id.twitch.tv/oauth2/token', data=payload)
        if response.status_code != 200:
            raise Exception(f'Failed to refresh token: {response.text}')
        
        data = response.json()
        self.OAuth_TOKEN = f"oauth:{data['access_token']}"
        self.REFRESH_TOKEN = data['refresh_token']

        self._update_env_file()
        self._update_runtime_env()
        
    def _update_env_file(self):
        with open(self.env_path, 'r', encoding='utf-8') as file:
            lines = file.readlines()

        with open(self.env_path, 'w', encoding='utf-8') as file:
            for line in lines:
                if line.startswith('TWITCH_OAUTH='):
                    file.write(f'TWITCH_OAUTH={self.OAuth_TOKEN}\n')
                elif line.startswith('TWITCH_REFRESH_TOKEN='):
                    file.write(f'TWITCH_REFRESH_TOKEN={self.REFRESH_TOKEN}\n')
                else:
                    file.write(line)

    def _update_runtime_env(self):
        if self.OAuth_TOKEN is not None:
            os.environ['TWITCH_OAUTH'] = self.OAuth_TOKEN
        if self.REFRESH_TOKEN is not None:
            os.environ['TWITCH_REFRESH_TOKEN'] = self.REFRESH_TOKEN

    def get_oauth_token(self):
        return self.OAuth_TOKEN


from dataclasses import dataclass
from typing import Any

@dataclass
class CommandInfo:
    command: str
    args: list[str]
    user: str

@dataclass
class CommandData:
    irc: Any
    channel: str
    command_info: CommandInfo


import socket

class IRCClient():
    def __init__(self, TOKEN: str, NICK: str, CHANNEL: str, SERVER: str = "irc.chat.twitch.tv", PORT: int = 6667):
        """
        Creates an IRC Client config.
        
        Args:
            TOKEN (str): The OAuth token (must start with 'oauth:')
            NICK (str): The bot's Twitch username
            CHANNEL (str): The channel to join (must start with '#')
            SERVER (str): IRC server hostname
            PORT (int): IRC port
        """
        
        if not TOKEN.startswith("oauth:"):
            raise ValueError("OAuth token must start with 'oauth:'")
        if not NICK:
            raise ValueError("Nick cannot be empty.")
        if not CHANNEL.startswith("#"):
            CHANNEL = f"#{CHANNEL}"
        
        self.token = TOKEN
        self.nickname = NICK
        self.channel = CHANNEL
        self.server = SERVER
        self.port = PORT
        self.irc = None
        self.recv_buffer = ""
    
    def connect(self):
        """
        Connects to the IRC Server and joins the specified channel.
        """
        
        self.irc = socket.socket()
        self.irc.connect((self.server, self.port))
        self.irc.settimeout(1)
        
        self.send_raw(f"PASS {self.token}")
        self.send_raw(f"NICK {self.nickname}")
        self.send_raw(f"JOIN {self.channel}")
        
    def send_raw(self, MESSAGE: str):
        """
        Sends a raw message to the IRC Server.
        
        Args:
            MESSAGE (str): The raw IRC message to send
        """
        if not self.irc:
            raise ConnectionError("IRC Connection is not established.")
        
        print(f"[IRC Raw] {MESSAGE}")
        self.irc.send((MESSAGE + "\r\n").encode("utf-8"))
    
    def recv(self, BUFFER_SIZE: int = 2048):
        """
        Receives data from the IRC Server.
        
        Args:
            BUFFER_SIZE (int): The buffer size for receiving data
        """
        if not self.irc:
            raise ConnectionError("IRC Connection is not established.")
        
        while '\r\n' not in self.recv_buffer:
            try:
                CHUNK = self.irc.recv(BUFFER_SIZE).decode("utf-8")
                if not CHUNK:
                    raise ConnectionError("Disconnected from server.")
                self.recv_buffer += CHUNK
            except socket.timeout:
                raise
            
        line, self.recv_buffer = self.recv_buffer.split('\r\n', 1)
        return line
    
    def close(self):
        """
        Closes the IRC Connection.
        """
        if self.irc:
            self.irc.close()
            self.irc = None


# From Imports
from .modules import IRCClient, Auth, cmd_handler, GetUsername, CommandData, CommandInfo
from dotenv import load_dotenv
from typing import cast

# Normal Imports
import os, threading, queue, socket

# <!-------< Main >-------!>
class Bot:
    def __init__(self, COMMANDS: dict ,ENV_Path = None, HANDLE: str = "!"):
        """
        Initialize the Twitch Bot
        
        Args:
            COMMANDS (dict): Example, {"add": add, "subtract": subtract}
            ENV_Path (str): While it's not required you should put your env file location so you can auto-update the OAuth details
            HANDLE (str): The command handle such as; '-help', '-hello', '-test'
        """
        load_dotenv()
        auth = Auth(ENV_PATH=ENV_Path)
        
        self.name = cast(str, os.getenv("BOTNAME"))
        self.token = cast(str, auth.get_oauth_token())
        self.channel = cast(str, os.getenv("CHANNEL"))
        
        self.handle = HANDLE
        self.commands = COMMANDS
        
        self.shutdown_requested = threading.Event()
        self.command_queue = queue.Queue()
        
        missing = [var for var, value in {
            "BOTNAME": self.name,
            "TOKEN": self.token,
            "CHANNEL": self.channel
        }.items() if not value]

        if missing:
            raise ValueError(f"Missing required environment variable(s): {', '.join(missing)}")
        
        self.irc = IRCClient(self.token, self.name, self.channel)
        self.irc.connect()
    
    def start(self):
        threading.Thread(
            target=self.worker,
            daemon=True
        ).start()
        self.main()
    
    def worker(self):
        while True:
            data = self.command_queue.get()
            
            if data is None:
                print("[WORKER] Shutdown signal received!")
                self.command_queue.task_done()
                break
            
            try:
                shutdown_request = cmd_handler(
                    data=data, 
                    COMMANDS=self.commands,
                    handle=self.handle
                )
                if shutdown_request:
                    self.shutdown_requested.set()
            except Exception as e:
                print(f"[WORKER] Error: {e}")
            finally:
                self.command_queue.task_done()
    
    def handle_message(self, RESPONSE: str):
        if "PRIVMSG" not in RESPONSE:
            return
        
        parts = RESPONSE.split(":", 2)
        if len(parts) < 3:
            return
        
        FULL_MSG = parts[2].strip()
        user = GetUsername(RESPONSE)
        words = FULL_MSG.split()
        
        if not words:
            return
        
        command = words[0].lower()
        args = words[1:]
        
        #self.command_queue.put((self.irc, self.channel, (command, args, user)))
        self.command_queue.put(
            CommandData(
                irc=self.irc,
                channel=self.channel,
                command_info=CommandInfo(
                    command=command,
                    args=args,
                    user=user
                )
            )
        )
    
    def main(self):
        print(f"[BOT] {self.name} is warming up...")
        
        try:
            while True:
                if self.shutdown_requested.is_set():
                    break
                
                try:
                    response = self.irc.recv()
                except socket.timeout as s:
                    if self.shutdown_requested.is_set():
                        print(f"[MAIN] Socket Timeout: {s}")
                        break
                    continue
                except Exception as e:
                    print(f"[Main] Error: {e}")
                    continue
                
                if response.startswith("PING"):
                    self.irc.send_raw("PONG :tmi.twitch.tv")
                else:
                    self.handle_message(response)
        finally:
            print(f"[BOT] Cleaning up after {self.name}...")
            self.command_queue.put(None)
            self.command_queue.join()
            self.irc.close()
            print(f"[BOT] Shutdown of {self.name} has been completed.")